@@ -82,6 +82,7 @@ module Text.XML.Stream.Parse
     , psDecodeIllegalCharacters
     , psRetainNamespaces
     , psEntityExpansionSizeLimit
+    , psIgnoreInternalEntityDeclarations
       -- *** Entity decoding
     , decodeXmlEntities
     , decodeHtmlEntities
@@ -147,6 +148,7 @@ import           Control.Monad.Trans.Class    (lift)
 import           Control.Monad.Trans.Maybe    (MaybeT (..))
 import           Control.Monad.Trans.Resource (MonadResource, MonadThrow (..),
                                                throwM)
+import           Data.Attoparsec.Internal     (concatReverse)
 import           Data.Attoparsec.Text         (Parser, anyChar, char, manyTill,
                                                skipWhile, string, takeWhile,
                                                takeWhile1, (<?>),
@@ -200,7 +202,7 @@ tokenToEvent ps es n (TokenBeginElement name as isClosed _) =
 
         addNS
             | not (psRetainNamespaces ps) && (isPrefixed || isUnprefixed) = id
-            | otherwise = (((tname, resolveEntities' ps es val):) .)
+            | otherwise = (. ((tname, resolveEntities' ps es val):))
           where
             resolveEntities' ps' es' xs =
               mapMaybe extractTokenContent
@@ -464,6 +466,10 @@ data ParseSettings = ParseSettings
     -- Default: @8192@
     --
     -- Since 1.9.1
+    , psIgnoreInternalEntityDeclarations :: Bool
+    -- ^ Whether to resolve any but the predefined entities.
+    --
+    -- Default: @False@
     }
 
 instance Default ParseSettings where
@@ -472,6 +478,7 @@ instance Default ParseSettings where
         , psRetainNamespaces = False
         , psDecodeIllegalCharacters = const Nothing
         , psEntityExpansionSizeLimit = 8192
+        , psIgnoreInternalEntityDeclarations = False
         }
 
 conduitToken :: MonadThrow m => ParseSettings -> ConduitT T.Text (PositionRange, Token) m ()
@@ -535,7 +542,10 @@ parseToken settings = do
                     do char' '['
                        ents <- parseDeclarations id
                        skipSpace
-                       return ents
+                       if psIgnoreInternalEntityDeclarations settings then
+                         return []
+                       else
+                         return ents
                   _ -> return []
         char' '>'
         newline <|> return ()
@@ -625,13 +635,15 @@ parseIdent = takeWhile1 valid <?> "identifier"
     valid '/'  = False
     valid ';'  = False
     valid '#'  = False
+    valid '['  = False
+    valid ']'  = False
     valid c    = not $ isXMLSpace c
 
 parseContent :: ParseSettings
              -> Bool -- break on double quote
              -> Bool -- break on single quote
              -> Parser Content
-parseContent (ParseSettings decodeEntities _ decodeIllegalCharacters _) breakDouble breakSingle = parseReference <|> parseTextContent where
+parseContent (ParseSettings decodeEntities _ decodeIllegalCharacters _ _) breakDouble breakSingle = parseReference <|> (parseTextContent <?> "text content") where
   parseReference = do
     char' '&'
     t <- parseEntityRef <|> parseHexCharRef <|> parseDecCharRef
@@ -659,11 +671,71 @@ parseContent (ParseSettings decodeEntities _ decodeIllegalCharacters _) breakDou
     case toValidXmlChar n <|> decodeIllegalCharacters n of
       Nothing -> fail "Invalid character from decimal character reference."
       Just c  -> return $ ContentText $ T.singleton c
-  parseTextContent = ContentText <$> takeWhile1 valid <?> "text content"
+
+  -- Turns @\r\n@ and @\r@ into @\n@. See
+  -- <https://www.w3.org/TR/REC-xml/#sec-line-ends>.
+  parseTextContent = do
+    -- Read until the end of this piece of content
+    -- OR until a carriage return. In the second case, we use
+    -- handleCR to normalize \r and \r\n into \n.
+    firstChunk <- takeWhile valid
+    mbC <- peekChar
+    case mbC of
+      Just '\r' ->
+        handleCR [firstChunk]
+      _ ->
+        exit firstChunk
+
+  -- This is a duplication of the logic above and could be used instead.
+  -- Specialising these cases to the case "full text content contains no carriage return"
+  -- considerably speeds up execution when no carriage returns are in the original source.
+  handleCRPeek chunks = do
+    mbC <- peekChar
+    case mbC of
+      Just '\r' ->
+        handleCR chunks
+      _ ->
+        exit' chunks
+
+  handleCR chunks = do
+    -- We know that the next character is a carriage return. Discard it.
+    _ <- anyChar
+    -- Read the next chunk.
+    chunk <- takeWhile valid
+    case T.uncons chunk of
+      -- If it starts with newline, we're good:
+      -- We've already discarded the carriage return.
+      -- This is the case that replaces \r\n by \n.
+      Just ('\n', _) ->
+        handleCRPeek $ chunk : chunks
+      -- Otherwise, we'll have to insert a newline.
+      -- This is the case that replaces \r by \n.
+      Just _ ->
+        handleCRPeek $ chunk : "\n" : chunks
+      -- If the chunk is empty, we've either hit another carriage
+      -- return or the end of this piece of content. Since we've discarded
+      -- a carriage return we need to insert a newline.
+      Nothing ->
+        handleCRPeek $ "\n" : chunks
+
+
+  -- exit and exit' fail if the emitted text content is empty.
+  -- exit' uses Data.Text.concat to efficiently concatenate the collected
+  -- chunks.
+  exit c
+    | T.null c = fail "parseTextContent"
+    | otherwise = pure $ ContentText c
+
+  exit' cs = exit $ T.concat $ reverse cs
+
+  -- Check whether a character is valid text content (e.g. not a <)
+  -- OR a carriage return. The latter is used above in parseTextContent
+  -- to normalize line endings.
   valid '"'  = not breakDouble
   valid '\'' = not breakSingle
   valid '&'  = False -- amp
   valid '<'  = False -- lt
+  valid '\r' = False
   valid _    = True
 
 -- | Is this codepoint a valid XML character? See
