@@ -53,7 +53,12 @@ main = hspec $ do
         it "conduit parser" testConduitParser
         it "can omit the XML declaration" omitXMLDeclaration
         it "doesn't hang on malformed entity declarations" malformedEntityDeclaration
+        it "escapes <>'\"& as necessary" caseEscapesAsNecessary
+        it "preserves the order of attributes" casePreservesAttrOrder
         context "correctly parses hexadecimal entities" hexEntityParsing
+        it "normalizes line endings" crlfToLfConversion
+        it "normalizes \\r at the end of a content" crlfToLfConversionCrAtEnd
+        it "normalizes multiple \\rs and \\r\\ns" crlfToLfConversionCrCrCr
     describe "XML Cursors" $ do
         it "has correct parent" cursorParent
         it "has correct ancestor" cursorAncestor
@@ -83,6 +88,7 @@ main = hspec $ do
         it "identifies unresolved entities" resolvedIdentifies
         it "decodeHtmlEntities" testHtmlEntities
         it "works for resolvable entities" resolvedAllGood
+        it "ignores custom entities when psResolveEntities is False" dontResolveEntities
         it "merges adjacent content nodes" resolvedMergeContent
         it "understands inline entity declarations" resolvedInline
         it "understands complex inline with markup" resolvedInlineComplex
@@ -558,6 +564,29 @@ malformedEntityDeclaration = do -- missing > after bim
                     _ -> False
         _ -> False
 
+caseEscapesAsNecessary :: Assertion
+caseEscapesAsNecessary = do
+    let doc = Res.Document (Res.Prologue [] Nothing [])
+                (Res.Element "a" (Map.fromList [("attr", "'<&val>'")])
+                    [Res.NodeContent "'\"<&test]]>\"'"])
+                []
+        result = Res.renderLBS def doc
+    result `shouldBe` "<?xml version=\"1.0\" encoding=\"UTF-8\"?><a attr=\"'&lt;&amp;val>'\">'\"&lt;&amp;test]]&gt;\"'</a>"
+
+casePreservesAttrOrder :: Assertion
+casePreservesAttrOrder = do
+    let doc = Document (Prologue [] Nothing [])
+                (Element "doc" [] [
+                  NodeElement (Element "el" [("attr1", [ContentText "1"]), ("attr2", [ContentText "2"])] []),
+                  NodeElement (Element "el" [("attr2", [ContentText "2"]), ("attr1", [ContentText "1"])] [])
+                ])
+                []
+        rendered = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><doc><el attr1=\"1\" attr2=\"2\"/><el attr2=\"2\" attr1=\"1\"/></doc>"
+        renderResult = D.renderLBS def doc
+        parseResult = D.parseLBS def rendered
+    renderResult `shouldBe` rendered
+    parseResult `shouldSatisfy` either (const False) (doc==)
+
 hexEntityParsing :: Spec
 hexEntityParsing = do
   it "rejects leading 0x" $
@@ -643,8 +672,8 @@ cursorParent, cursorAncestor, cursorOrSelf, cursorPreceding, cursorFollowing,
     cursorAnyElement, cursorElement, cursorLaxElement, cursorContent,
     cursorAttribute, cursorLaxAttribute, cursorHasAttribute,
     cursorAttributeIs, cursorDeep, cursorForce, cursorForceM,
-    resolvedIdentifies, resolvedAllGood, resolvedMergeContent,
-    testHtmlEntities
+    resolvedIdentifies, resolvedAllGood, dontResolveEntities,
+    resolvedMergeContent, testHtmlEntities
     :: Assertion
 cursorParent = name (Cu.parent bar2) @?= ["foo"]
 cursorAncestor = name (Cu.ancestor baz2) @?= ["bar2", "foo"]
@@ -721,6 +750,25 @@ resolvedAllGood =
   where
     xml = "<foo><bar/><baz/></foo>"
 
+dontResolveEntities =
+  D.parseLBS_ settings xml @=? expectedDocument
+  where
+    settings = def { P.psIgnoreInternalEntityDeclarations = True }
+    xml = mconcat
+      [ "<!DOCTYPE mydt [ <!ENTITY foo \"fooby\" > ]>"
+      , "<root>&gt;&foo;&bar;</root>"
+      ]
+
+    expectedDocument =
+      Document
+        (Prologue [] (Just (Doctype "mydt" Nothing)) [])
+        (Element "root" mempty
+          [ NodeContent (ContentText ">")
+          , NodeContent (ContentEntity "foo")
+          , NodeContent (ContentEntity "bar")
+          ])
+        []
+
 resolvedMergeContent =
     Res.documentRoot (Res.parseLBS_ def xml) @=?
     Res.Element "foo" Map.empty [Res.NodeContent "bar&baz"]
@@ -750,7 +798,7 @@ testRenderComments =do
 
 resolvedInline :: Assertion
 resolvedInline = do
-    Res.Document _ root _ <- return $ Res.parseLBS_ Res.def "<!DOCTYPE foo [<!ENTITY bar \"baz\">]><foo>&bar;</foo>"
+    Res.Document _ root _ <- return $ Res.parseLBS_ Res.def "<!DOCTYPE foo[<!ENTITY bar \"baz\">]><foo>&bar;</foo>"
     root @?= Res.Element "foo" Map.empty [Res.NodeContent "baz"]
     Res.Document _ root2 _ <- return $ Res.parseLBS_ Res.def "<!DOCTYPE foo [<!ENTITY bar \"baz\">]><foo bar='&bar;'/>"
     root2 @?= Res.Element "foo" (Map.singleton "bar" "baz") []
@@ -1041,3 +1089,21 @@ caseEscapesCDATA = do
                 []
         result = Res.renderLBS (def { Res.rsUseCDATA = const True }) doc
     result `shouldBe` "<?xml version=\"1.0\" encoding=\"UTF-8\"?><a><![CDATA[]]]]><![CDATA[>]]></a>"
+
+crlfToLfConversion :: Assertion
+crlfToLfConversion = (elementContent $ documentRoot crlfDoc) `shouldBe` crlfContent
+    where
+        crlfDoc = D.parseLBS_ def "<crlf>Hello,\rWorld!\r\nWe don't like your kind of line endings around here.\r\n</crlf>"
+        crlfContent = [ContentText "Hello,\nWorld!\nWe don't like your kind of line endings around here.\n"]
+
+crlfToLfConversionCrAtEnd :: Assertion
+crlfToLfConversionCrAtEnd = (elementContent $ documentRoot doc) `shouldBe` content
+    where
+        doc = D.parseLBS_ def "<crlf>Hello, World!\r</crlf>"
+        content = [ContentText "Hello, World!\n"]
+
+crlfToLfConversionCrCrCr :: Assertion
+crlfToLfConversionCrCrCr = (elementContent $ documentRoot doc) `shouldBe` content
+    where
+        doc = D.parseLBS_ def "<crlf>\r\r\r\n\r\r\r</crlf>"
+        content = [ContentText "\n\n\n\n\n\n"]
